/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
 
package au.com.terabit.jproactor;

//import java.nio.channels.ServerSocketChannel;
//import java.nio.channels.SocketChannel;

/**
 * Th <code>OpAccept</code> class implements <i>accept</i> IO operation.
 * <p>
 * Copyright &copy; 2003 Terabit Pty Ltd. All rights reserved.
 *
 * @author <a href="mailto:libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class OpAccept extends IOOperation
{
    
    /** Newly accepted connection instance */
    protected AsynchChannel m_acceptedChannel = null;

    /** Operation's handler */
    private AsynchAcceptHandler m_handler = null;


    /**
     * Creates accept operation using specified <code>TCPAsynchChannel</code>.  
     * 
     * @param achannel <code>TCPAsynchChannel</code> instance for this accept operation
     */
    OpAccept( AsynchChannel achannel, AsynchAcceptHandler handler )
    {
        super( IOOperation.OP_ACCEPT, achannel );
        
        this.m_state = IOOperation.STATE_STARTED;
        this.m_handler = handler;
    }

    protected boolean execute ()
    {
       try
       {
           m_acceptedChannel = m_asynchChannel.finishAccept (this);
        
           if ( m_acceptedChannel != null)
           {                       
               m_state = IOOperation.STATE_FINISHED;
           }
       }
       catch (Exception e)
       {
           m_state = IOOperation.STATE_FINISHED;
           m_errorCause = e;
       }
   
       if (m_state == IOOperation.STATE_FINISHED)
       {
           ++m_asynchChannel.m_readInfo.m_opFinished;
           return true;
       }
       return false;
         
    }
    
    /**
     * Dispatches accepted channel to the associated protocol so it can start
     * execution.
     * 
     * @see IOOperation#onComplete()
     * @see AsynchChannelHandler#acceptCompleted(OpAccept)
     * @see 
     */
    public void onComplete()
    {
        try
        {
            if ( m_handler != null )
            {
                m_handler.acceptCompleted( this );
            }
        }
        catch ( Exception e )
        {
            System.out.println( "OpAccept::onComplete:" + e.getMessage() );
        }
    }
    
    /**
     * Returns the result of the accept operation.
     * 
     * @return a newly accepted socket channel
     */
    public AsynchChannel getAcceptedChannel() { return m_acceptedChannel; }

}
