/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.jproactor;
import java.net.SocketAddress;


/**
 * The <tt>connect</tt> operation implementation.
 *
 * @author <a href="mailto:libman@terabit.comg">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public class OpConnect extends IOOperation
{  
	protected SocketAddress m_remoteAddr = null;

    /** Operation's handler */
    private AsynchConnectHandler m_handler = null;


    /* ---------------------------------------------------------------------- */
    
	OpConnect( AsynchChannel channel,
               AsynchConnectHandler handler,
               SocketAddress addr ) throws Exception
	{
		super( IOOperation.OP_CONNECT, channel);
        
        this.m_handler = handler;
		this.m_remoteAddr = addr;
	}
    
    protected boolean execute ()
    {
       try
       {
           boolean rc = m_asynchChannel.finishConnect (this);
        
           if (rc)
           {                       
               m_state = IOOperation.STATE_FINISHED;
           }
       }
       catch (Exception e)
       {
           m_state = IOOperation.STATE_FINISHED;
           m_errorCause = e;
       }
   
       if (m_state == IOOperation.STATE_FINISHED)
       {
           ++m_asynchChannel.m_writeInfo.m_opFinished;
           return true;
       }
       return false;
            
    }
    
	public void onComplete() 
	{
		try
		{
            if (m_handler != null)
            {
                m_handler.connectCompleted( this );
            }
		}
		catch (Exception e)
		{
			System.out.println ("OpConnect::onComplete: " + e.getMessage());
		}
	}

    /**
     * Tells whether connect operation on the channel has successfully completed
     * or not.
     * 
     * @return <tt>true</tt> if associated asynchronous chanel is connected,
     *         <tt>false</tt> otherwise 
     */
    public boolean isConnected()
    {
        return m_state == STATE_FINISHED && getError() == null;
    }
}
