/*
 *********************************************************************
 * Created on 6/01/2006
 *
 * Copyright (C) 2003-06 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.jproactor;

/**
 * This class provides implementation of the timer operations.
 *
 * @author <a href="mailto:libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 */
public final class OpTimer
{

    /** Timer state values */
    public enum State { CREATED, QUEUED, RUNNING, CANCELLED, EXPIRED }
    
    /**
     * Timer expiration time: number of milliseconds since 1 Jan 1970 00:00:00 GMT
     */
    public final long m_expiryTime; 
        
    /** Timer's handler instance to dispatch timer expiry event. */
    final private AsynchTimerHandler m_handler;

    ///** IOOperation for which this timee was created, optional */
    //private IOOperation m_operation = null;
    
    /** Operation's state */
    private State m_state = State.CREATED;

    private volatile Object attachment = null;  


    // ========================================================================
    //

    /**
     * 
     * @param handler the timer notification receiver
     * @param delay time delay from now when to expire the timer, milliseconds
     */
    public OpTimer(long delay, AsynchTimerHandler handler )
    {
        this.m_handler = handler;
        this.m_expiryTime = System.currentTimeMillis() + delay;
    }

    /**
     * Called when timer has expired - notifies protocol instance.
     * 
     * @see IOOperation#onComplete()
     */
    public void onComplete()
    {
            //????
            if ( m_state == State.CANCELLED ) return;
            
            m_state = State.RUNNING;

            try
            {
                if (m_handler != null)
                {
                    m_handler.timerExpired( this );
                }
            }
            catch( Exception e )
            {
                System.out.println ("OpTimer::onComplete: " + e.getMessage());
            }
            m_state = State.EXPIRED;
    };

    /** Cancels this timer */
    public void    cancel()      { m_state = State.CANCELLED; }
    public boolean isCancelled() { return m_state == State.CANCELLED; }

    void setState( OpTimer.State state ) { m_state = state; }
    
    /**
     * Attaches an object to this operation.
     * <p>
     * Attachments can be useful when executig {@link #onComplete()} operation 
     * and be retrieved using {@link #attachment()} operation. 
     *  
     * @param value object to be attached, can be a <tt>null</tt> value
     * @return The previously-attached object, if any, otherwise <tt>null</tt>
     */
    public final Object attach( Object value )
    {
        Object old = this.attachment;
        this.attachment = value;

        return old;
    }
    
    /**
     * Returns currently attached object.
     * 
     * @return attachement
     */
    public final Object attachment() {  return this.attachment; }

}
