/*
 *********************************************************************
 * Created on ${date}
 *
 * Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
 *
 * This file may be distributed and used only under the terms of the  
 * Terabit Public License as defined by Terabit Pty Ltd of Australia   
 * and appearing in the file tlicense.txt included in the packaging of
 * this module and available at http://www.terabit.com.au/license.php.
 *
 * Contact support@terabit.com.au for any information
 *********************************************************************
 */
package au.com.terabit.jproactor;

//import java.net.SocketAddress;
//import java.nio.ByteBuffer;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;

//import java.util.concurrent.*;
//import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.atomic.AtomicBoolean;
//import java.util.concurrent.atomic.AtomicInteger;
//import java.util.concurrent.atomic.AtomicReference;

//import java.util.LinkedList;


/**
 * The <code>TCPAsynchChannel</code> is an intermidiate interface that serves as
 * mediator between <code>Demultiplexor</code> and <code>AsynchHandler</code>
 * <p>
 * All implemented protocols should be based on this interface.
 *
 * @author <a href="mailto:libman@terabit.com.au">Yevgeny Libman</a>
 * @version <code>$Revision$ $Date$</code>
 * 
 * @see IOOperation
 */
final public class TCPAsynchChannel extends AsynchChannel
{

    /* ---------------------------------------------------------------------- */
    //
    // instance variables
    private boolean m_flgOnce =  
        (Demultiplexor.dispatchStategy_ == Demultiplexor.STRATEGY_BALANCE);
 
    /** multiplexor reference */
    private Demultiplexor m_demultiplexor = null; 
    
    
    /** m_asynchChannel's channel */
    private SelectableChannel m_selChannel = null;
    
    
    /** dispatch m_state can be DISP_NONE, DISP_QUEUED, DISP_ACTIVE */
    private AtomicBoolean  m_dispatchActive = 
    	new AtomicBoolean(false); 
    
    /** ready events as reported by selector */
    private int m_readyEvents = 0;
    
    /** event mask that is currently active in selector */
    private int m_remainEvents = 0;
    
    /** new mask that should be activated in selector after dispatch operation */
    private int m_interestEvents = 0;
    
    

    /* ---------------------------------------------------------------------- */
    
    /**
     * Creates <code>TCPAsynchChannel</code> instance given specified 
     * <code>Demultiplexor</code>, <code>AsynchHandler</code> and 
     * <code>SelectableChannel</code> instances.
     *
     * @param m <code>Demultiplexor</code> instance this m_asynchChannel works with
     * @param p <code>AsynchHandler</code> protocol instance
     * @param sc <code>SelectableChannel</code> instance this m_asynchChannel operates on
     */
    public TCPAsynchChannel( Demultiplexor m, SelectableChannel sc )
    {
        //super (handler);
        super();
        this.m_demultiplexor = m;
        this.m_selChannel    = sc;
        
        m_demultiplexor.declareInterest (this);
    }

 
    /**
     * Completion dispather for this m_asynchChannel.
     * This method is called from Demultiplexor thread pool when
     * m_asynchChannel has completed operations.
     * 
     * It extacts all completioned operations from the completion queue
     * assosiated with this {@link #m_asynchChannel} and performs  upcalls
     * to the associated ProtocolHandlers
     */
    final void dispatchEvents (int readyEvents, int remainEvents)
    {
        boolean flgInterest = false;
        
        m_lock.lock();
    	boolean activeFlg = m_dispatchActive.compareAndSet(false,true);
    	try
    	{
    		m_readyEvents |= readyEvents;
    		m_remainEvents = remainEvents;
    		m_interestEvents &= ~readyEvents;
    		
    	   	if (!activeFlg)
        		return;
            
            // execute each IO operation only once then update interested events
            // mask it's possible that during IOOperation::onComplete() new
            // operations will be added
            for (; ;)
            {
                IOOperation op = null;
            
                if ((m_readyEvents & (SelectionKey.OP_ACCEPT | SelectionKey.OP_READ)) !=0)
                {
                    op = executeListAndDispatch (m_readQue);
                    if (op == null)
                    {
                        m_interestEvents &= ~(SelectionKey.OP_ACCEPT | SelectionKey.OP_READ);
                    }
                    else
                    {
                        m_interestEvents |= op.m_type;
                        m_readyEvents &= ~op.m_type;
                    }
                }
                                
                if ((m_readyEvents & (SelectionKey.OP_CONNECT | SelectionKey.OP_WRITE)) !=0)
                {
                    op = executeListAndDispatch (m_writeQue);
                    if (op == null)
                    {
                        m_interestEvents &= ~(SelectionKey.OP_CONNECT | SelectionKey.OP_WRITE);
                    }
                    else
                    {
                        m_interestEvents |= op.m_type;
                        m_readyEvents &= ~op.m_type;
                    }
                }
                
                if (m_flgOnce || (m_interestEvents & m_readyEvents ) == 0)
                {
                    break;
                }
            }
            
            if (!checkForClose())
            {
                 if (m_interestEvents != 0 &&
                     m_interestEvents != m_remainEvents ) // YL
    		     {
    		         flgInterest = true;
    		     }
            }
            
    	}
    	catch (Exception e)
    	{
    		e.printStackTrace();
    	}
    	finally
    	{
    		m_lock.unlock();
            
    	}
    	
        m_dispatchActive.set(false);
        
        if (flgInterest)
        {
            m_demultiplexor.declareInterest(this);
        }
}

    
 
    protected int finishRead(OpRead op) throws Exception
    {
        //System.out.println( Thread.currentThread().getName() + ": Executing Read IO" );

        SocketChannel channel = (SocketChannel) m_selChannel;
        
        int n = channel.read ( op.getBuffer() );

        return n;
    }
    
    /**
     * Writes data into associated channel, when channel is ready for
     * non-blocking write.
     *
     * @see IOOperation#execute()
     */
    protected int finishWrite(OpWrite op) throws Exception
    {
        //System.out.println( Thread.currentThread().getName() + ": Executing Write IO" );
    
        SocketChannel channel = (SocketChannel) m_selChannel;
    
        int  n = channel.write(op.getBuffer () );
        
        return n;
    }
    
    protected AsynchChannel finishAccept(OpAccept op) throws Exception
    {
        //System.out.println( Thread.currentThread().getName() + ": Executing Accept IO" );

        // get server socket channel from m_asynchChannel
        ServerSocketChannel server = (ServerSocketChannel) m_selChannel;
        SocketChannel acceptedChannel = server.accept();
            
        // check that accept actually has happened 
        if (acceptedChannel != null )
        {
            acceptedChannel.socket().setReuseAddress(true);
            
            TCPAsynchChannel asynchChannel = 
                new TCPAsynchChannel( m_demultiplexor, acceptedChannel);
            
            return asynchChannel;
        }
            
        return null;
    }
    
    protected boolean finishConnect (OpConnect op) throws Exception
    {
        //System.out.println( Thread.currentThread().getName() + ": Executing Read IO" );

        SocketChannel channel = (SocketChannel) m_selChannel;
        
        if (channel.finishConnect())
        {
            setConnected ();
            
            return true;
        }
        return false;
    }

    public boolean startClose ()
    {
        m_demultiplexor.declareInterest(this);
        return true;
    }
    
    
    protected void startRead(OpRead op) throws Exception
    {
        boolean flgInterest = setInterest (op.m_type);
        if (flgInterest)   
        {
            m_demultiplexor.declareInterest(this);
        }
    }
    
    protected void startWrite(OpWrite op) throws Exception
    {
        boolean flgInterest = setInterest (op.m_type);
        if (flgInterest)   
        {
            m_demultiplexor.declareInterest(this);
        }
    }
    
    protected void startAccept(OpAccept op) throws Exception
    {
        boolean flgInterest = setInterest (op.m_type);
        if (flgInterest)   
        {
            m_demultiplexor.declareInterest(this);
        }
    }
    
    protected void startConnect(OpConnect op) throws Exception
    {
        SocketChannel channel = (SocketChannel) m_selChannel;
        channel.configureBlocking(false);
        
        int mask = op.m_type;
        
        if (channel.connect(op.m_remoteAddr))
        {
            op.m_state = IOOperation.STATE_FINISHED;
            this.setConnected ();
            mask  |= SelectionKey.OP_WRITE;
        }
        else
        {
            op.m_state = IOOperation.STATE_STARTED;
        }

        boolean flgInterest = setInterest (mask);
        if (flgInterest)   
        {
            m_demultiplexor.declareInterest(this);
        }
        
    }
    
    private boolean setInterest (int  opType)
    {
        m_interestEvents |= opType;
        
        if ((m_remainEvents & opType) != 0)
            return false;
        
        if (m_dispatchActive.get())
            return false;
        
        // need to wake up selector
        return true;
    }
    
    /**
     * Schedules timeout action to be called back on the associated 
     * <code>AsynchHandler</code> instance.
     * <p>
     * Note: while IO operations will be delivered to the handler in a
     * serialized way, timer notifications might be delivered concurrently.    
     * 
     * @param delay delay in milliseconds before timeout action is to be executed. 
     * @return OpTimer operation - future result
     */
    public void startTimer(OpTimer op)
    {
        m_demultiplexor.startTimer( op );
    }

 
    // -----------------------------------------------------------------------
    //
    // Package visible methods
    
    /** 
     * Returns Selectable channel associated 
     * with this Asynch channel.
     */
    SelectableChannel getSelectableChannel() 
    { 
    	return m_selChannel;
    }
    
    /**
     * updateInterest is called from Demultiplexor
     * in LEADER_STATE, i.e. with locked leader lock
     * @param  selector to declare/cancel m_asynchChannel interest
     * @return true  - if interesrt updated
     *         false - failure, m_Channel is closed and must be dispatched
     *  
     */
    boolean updateInterest (Selector selector)
    {
     	boolean rc = false;
        m_lock.lock();
    	try
        {
        	if (isOpened ()) 
        	{
    	       	m_remainEvents |= m_interestEvents;
	            m_interestEvents = 0;
	 			SelectionKey key = m_selChannel.keyFor (selector);
		
				if (key == null)
				{
					m_selChannel.configureBlocking( false );
					key = m_selChannel.register (selector, m_remainEvents, this);
				}
				else
				{
					key.interestOps (m_remainEvents);
				}
				rc = true;
			}
        }
        catch ( Exception e ) 
        { 
        	e.printStackTrace();
        	rc = false;
        }
        finally
        {
           	if (!rc)
        	{
        	   closeInternal ();
        	}
        	m_lock.unlock ();
        }
        	
        return rc;
    }
    
    boolean onSelectorClosed (Selector selector)
    {
        m_lock.lock();
        try
        {
            setClosing ();
            closeInternal ();
        }
        finally
        {
            m_lock.unlock ();
        }
        return true;
    }
    
    private void closeInternal ()
    {
        // this will automatically cancel all keys
        try
        {
            if (m_selChannel != null)
            {
                m_selChannel.close();
            }
        }
        catch (Exception xc)
        {
            xc.printStackTrace();
        }
        finally
        {
            m_selChannel = null;
            m_remainEvents = IOOperation.OP_ALL;
            m_interestEvents = 0;
        }
    }
  
    
}